/******************************************************************************
* File Name: main.c
*
* Version: 1.10
*
* Description: This project showcases the capability of EZ-BLE PRoC Module to
* communicate with a BLE Central device. The data communicated is CapSense slider
* centroid and button status sent from EZ-BLE PRoC Module over a custom service.
* This custom service allows notifications to be sent to central device after
* notifications are enabled. This project receives CapSense slider centroid and
* button status via I2C interface from PSoC 4000S device. The BLE central device
* can be either BLE capable smartphone with CySmart mobile application or CySmart
* PC tool. This project also includes low power mode operation, ideal for battery
* operated devices. The project utilizes Deep Sleep feature of both BLESS and CPU
* to remain in low power mode as much as possible, while maintaining the BLE
* connection and  data transfer. This allows the device to run on coin cell battery
* for long time.
*
* Related Document: CE210709 CapSense Linear Slider and Buttons.pdf
*
* Hardware Dependency: See CE210709 CapSense Linear Slider and Buttons.pdf
*
*******************************************************************************
* Copyright (2018), Cypress Semiconductor Corporation. All rights reserved.
*******************************************************************************
* This software, including source code, documentation and related materials
* (“Software”), is owned by Cypress Semiconductor Corporation or one of its
* subsidiaries (“Cypress”) and is protected by and subject to worldwide patent
* protection (United States and foreign), United States copyright laws and
* international treaty provisions. Therefore, you may use this Software only
* as provided in the license agreement accompanying the software package from
* which you obtained this Software (“EULA”).
*
* If no EULA applies, Cypress hereby grants you a personal, nonexclusive,
* non-transferable license to copy, modify, and compile the Software source
* code solely for use in connection with Cypress’s integrated circuit products.
* Any reproduction, modification, translation, compilation, or representation
* of this Software except as specified above is prohibited without the express
* written permission of Cypress.
*
* Disclaimer: THIS SOFTWARE IS PROVIDED AS-IS, WITH NO WARRANTY OF ANY KIND, 
* EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, NONINFRINGEMENT, IMPLIED 
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. Cypress 
* reserves the right to make changes to the Software without notice. Cypress 
* does not assume any liability arising out of the application or use of the 
* Software or any product or circuit described in the Software. Cypress does 
* not authorize its products for use in any products where a malfunction or 
* failure of the Cypress product may reasonably be expected to result in 
* significant property damage, injury or death (“High Risk Product”). By 
* including Cypress’s product in a High Risk Product, the manufacturer of such 
* system or application assumes all risk of such use and in doing so agrees to 
* indemnify Cypress against all liability.
*******************************************************************************/
/*******************************************************************************
*   Included Headers
*******************************************************************************/

#include <project.h>
#include "BLEApplications.h"
#include "HandleLowPower.h"

/* I2C read buffer size */
#define I2C_BUF_SIZE		        (4u)

/* I2C slave address */
#define I2C_SLAVE_ADDRESS           (0x08u)

/* Sets the boundary between the read/write and read only areas.
*  The read/write area is first, followed by the read only area. */
#define I2C_RW_SIZE                 (0u)

/* EZI2C address offset */
#define I2C_ADDRESS_OFFSET          (0u)

/****************************************************************************/
/*'deviceConnected' flag is used by application to know whether a Central device  
* has been connected. This is updated in BLE event callback */
extern uint8 deviceConnected;

/*'startNotification' flag is set when the central device writes to CCC (Client  
* Characteristic Configuration) of the CapSense characteristic to 
* enable notifications */
extern uint8 startNotification;	

/* 'restartAdvertisement' flag is used to restart advertisement */
extern uint8 restartAdvertisement;

/* I2C buffer for storing the data read from I2C slave device */
uint8 i2cBuffer[] = {0, 0, 0, 0};

/**************************Function Declarations*****************************/
void InitializeSystem(void);
void HandleCapSense(void);

/*******************************************************************************
* Function Name: main
********************************************************************************
* Summary:
*        System entrance point. This calls the BLE routine functions as well as
* function for handling CapSense Slider and Buttons changes
*
* Parameters:
*  void
*
* Return:
*  int
*

*******************************************************************************/
int main()
{
    /* This function will initialize the system resources such as BLE and I2C */
	InitializeSystem();
	
    for(;;)
    {
		/*Process Event callback to handle BLE events. The events generated and 
		* used for this application are inside the 'CustomEventHandler' routine*/
		CyBle_ProcessEvents();
		
		/* Handle Slider and Buttons data and CCCD value update only if BLE device is connected */
		if(TRUE == deviceConnected) 
		{
			/* After the connection, send new connection parameter to the Client device 
			* to run the BLE communication on desired interval. This affects the data rate 
			* and power consumption. High connection interval will have lower data rate but 
			* lower power consumption. Low connection interval will have higher data rate at
			* expense of higher power. This function is called only once per connection. */
			UpdateConnectionParam();					
			
			if(CYBLE_BLESS_STATE_ECO_STABLE ==CyBle_GetBleSsState())
			{
				/*Check for CapSense data change and report to BLE central device*/
				HandleCapSense();
			}
		}

		if(restartAdvertisement)
		{
			/* Reset 'restartAdvertisement' flag*/
			restartAdvertisement = FALSE;

			/* Start Advertisement and enter Discoverable mode*/
			CyBle_GappStartAdvertisement(CYBLE_ADVERTISING_FAST);	
			
			/* Process the new event */
			CyBle_ProcessEvents();
		}
		
		#ifdef ENABLE_LOW_POWER_MODE
			/* Put system to Deep sleep, including BLESS, and wakeup on interrupt. 
			* The source of the interrupt can be either BLESS Link Layer in case of 
			* BLE advertisement and connection or by User Button press during BLE 
			* disconnection */
			HandleLowPowerMode();
		#endif
    }	/* This is end of for(;;) */
}

/*******************************************************************************
* Function Name: InitializeSystem
********************************************************************************
* Summary:
*        Starts the components used in the project
*
* Parameters:
*  void
*
* Return:
*  void
*
*******************************************************************************/
void InitializeSystem(void)
{
	/* Enable global interrupt mask */
	CyGlobalIntEnable;

	/* Start BLE component and register the CustomEventHandler function. This 
	* function exposes the events from BLE component for application use */
	CyBle_Start(CustomEventHandler);			
	
    /* Begin I2C master component operation */
	I2CHW_Start();    
}

/*******************************************************************************
* Function Name: HandleCapSense
********************************************************************************
* Summary:
*       Read the Slider and Buttons data from the sensor and update the BLE
*		custom notification value.
*
* Parameters:
*  void
*
* Return:
*  void
*
*******************************************************************************/
void HandleCapSense(void)
{
	static uint8 sliderValue = 0;
    static uint8 prevSliderVaule = 0;
	static uint8 buttonValue[3]; 
    static uint8 prevButtonStat = 0;
    
    /* Read entire data buffer from the slave device */
    I2CHW_I2CMasterReadBuf(I2C_SLAVE_ADDRESS, i2cBuffer, I2C_BUF_SIZE, 
                    I2CHW_I2C_MODE_COMPLETE_XFER);    
    while (0u == (I2CHW_I2CMasterStatus() & I2CHW_I2C_MSTAT_RD_CMPLT))
    {
        CyBle_ProcessEvents();
    }

	if(startNotification & SLIDER_CCCD_NTF_BIT_MASK)
	{			
        sliderValue = i2cBuffer[0];
        if(prevSliderVaule != sliderValue)
        {
            SendDataOverCapSenseSliderNotification(sliderValue);
        }
        prevSliderVaule = sliderValue;
	}
	
	if(startNotification & BUTTON_CCCD_NTF_BIT_MASK)
	{
		/* See this link for more details: http://www.cypress.com/file/175556/download */
		
	    buttonValue[0] = i2cBuffer[1];
        buttonValue[1] = i2cBuffer[2];
        buttonValue[2] = i2cBuffer[3];
        if(prevButtonStat != buttonValue[1])
        {
		    SendDataOverCapSenseButtonNotification(buttonValue);
        }
        prevButtonStat = buttonValue[1];
	}		
}

/* [] END OF FILE */
